/* Robust Improved complex divide Baudin and Smith */
/*
  The robust version of the Baudin and Smith improved method
  with improvements (RMIN, RBIG) by Patrick McGehearty
 */

#include <stdio.h>
#include <math.h>
#include <complex.h>
#include <float.h>
#define RBIG LDBL_MAX/2
#define RMIN LDBL_MIN
#define RMIN2 LDBL_EPSILON
#define RMINSCAL 1/LDBL_EPSILON

long double complex cdivld(long double complex x, long double complex y)
{
  long double a, b, c, d, ratio, denom, xx, yy;
  long double ax, bx, cx, dx;
  long double complex res;

  a = creall(x);
  b = cimagl(x);
  c = creall(y);
  d = cimagl(y);
  if (fabsl(c) < fabsl(d))
    {
      if (fabsl(d) >= RBIG)
	{
	  a = a * 0.5;
	  b = b * 0.5;
	  c = c * 0.5;
	  d = d * 0.5;
	}
      /*
	If max(c,d) < 2^-512, we can scale up by 2^510, to save
	some underflows without creating new overflows in final result.
       */
      if (fabsl(d) < RMIN2)
	{
	  a = a * RMINSCAL;
	  b = b * RMINSCAL;
	  c = c * RMINSCAL;
	  d = d * RMINSCAL;
	}
      ratio = c / d;
      denom = d + c * ratio;
      if (fabsl(ratio) > RMIN) {
	xx = (b + a * ratio) / denom;
	yy = (b * ratio - a) / denom;
      } else {
	xx = (b + c * (a / d)) / denom;
	yy = (c * (b / d) - a) / denom;
      }
    } 
  else
    {
      if (fabsl(c) >= RBIG)
	{
	  a = a * 0.5;
	  b = b * 0.5;
	  c = c * 0.5;
	  d = d * 0.5;
	}
      /*
	If max(c,d) < 2^-512, we can scale up by 2^510, to save
	some underflows without creating new overflows in final result.
       */
      if (fabsl(c) < RMIN2)
	{
	  a = a * RMINSCAL;
	  b = b * RMINSCAL;
	  c = c * RMINSCAL;
	  d = d * RMINSCAL;
	}
      ratio = d / c;
      denom = c + d * ratio;
      if (fabsl(ratio) > RMIN)
	{
	  xx = (a + b * ratio) / denom;
	  yy = (b - a * ratio) / denom;
	} 
      else
	{
	  xx = (a + d * (b / c)) / denom;
	  yy = (b - d * (a / c)) / denom;
	}
    }

  /* Recover infinities and zeros that computed as NaN+iNaN; the only cases
     are nonzero/zero, infinite/finite, and finite/infinite.  */
  if (isnan (xx) && isnan (yy))
    {
      ax = creal(x);
      bx = cimag(x);
      cx = creal(y);
      dx = cimag(y);
      if (denom == 0.0 && (!isnan (ax) || !isnan (bx)))
	{
	  xx = copysignl (INFINITY, cx) * ax;
	  yy = copysignl (INFINITY, cx) * bx;
	}
      else if ((isinf (ax) || isinf (bx))
	       && isfinite (cx) && isfinite (dx))
	{
	  ax = copysignl (isinf (ax) ? 1 : 0, a);
	  bx = copysignl (isinf (bx) ? 1 : 0, b);
	  xx = INFINITY * (ax * cx + bx * dx);
	  yy = INFINITY * (bx * cx - ax * dx);
	}
      else if ((isinf (c) || isinf (d))
	       && isfinite (a) && isfinite (b))
	{
	  cx = copysignl (isinf (cx) ? 1 : 0, c);
	  dx = copysignl (isinf (dx) ? 1 : 0, d);
	  xx = 0.0 * (ax * cx + bx * dx);
	  yy = 0.0 * (bx * cx - ax * dx);
	}
      else
	{
	  xx = 0.0/0.0;
	  yy = 0.0/0.0;
	}
    }
  res = (long double complex)(xx + yy * I);
  return res;
}
