/*
  Build with gcc -lm -O2 chkf_cdiv.c cdivf_t?.c -o chkf_t?
  Program to test float complex divide 
  for correct results on selected values.
  Checking known failure points.
*/

#include <math.h>
#include <complex.h>
#include <stdio.h>
extern void abort();
extern void exit();
extern float complex cdivf(float complex a,float complex b);

#define SMALL (0x1.0p-126)
#define MAXBIT 24
#define ERRLIM 6
int debug = 1;

/*
  Compare c (computed value) with z (expected value)
  Return 0 if within allowed range
  Return 1 if not
*/
int match(float complex c, float complex z)
{
  float rz, iz, rc, ic;
  float rerr, ierr, rmax;
  float biterr;
  rz = __real__ z;
  iz = __imag__ z;
  rc = __real__ c;
  ic = __imag__ c;
  if (isnan(rc) || isnan(ic)) return 1;

  if (fabsf(rz) > SMALL)
    {
      rerr = fabsf(rz-rc)/fabsf(rz);
    }
  else if (fabsf(rz) == 0.0)
    {
      rerr = fabsf(rc);
    }
  else
    {
      rerr = fabsf(rz-rc)/SMALL;
    }

  if (fabsf(iz) > SMALL)
    {
      ierr = fabsf(iz-ic)/fabsf(iz);
    }
  else if (fabsf(iz) == 0.0)
    {
      ierr = fabsf(ic);
    }
  else
    {
      ierr = fabsf(iz-ic)/SMALL;
    }
  rmax = fmax(rerr, ierr);
  biterr = 0;
  if ( rmax != 0.0)      
    {
      biterr = log2(rmax) + MAXBIT + 1;
    }
  if (debug)
      {
	/* debug */
	printf("biterr =%13.1f\n", biterr);
	printf("expect  zr %-16.6a zi   =%-16.6a\n",rz,iz);
	printf("observe cr %-16.6a ci   =%-16.6a\n\n",rc,ic);
      }

  if (biterr >= ERRLIM)
    return 0;
  else
    return 1;
}


int main(int argc, char** argv)
{
  float complex a,b,c,z;
  float xr[4], xi[4], yr[4], yi[4], zr[4], zi[4];
  float cr, ci;
  int i;
  int ok = 1;
  xr[0] = 0x1.0b1600p-133;
  xi[0] = 0x1.5e1c28p+54;
  yr[0] = -0x1.cdec8cp-119;
  yi[0] = 0x1.1e72ccp+32;
  zr[0] = 0x1.38e502p+22;
  zi[0] = -0x1.f89220p-129;

  xr[1] = -0x1.b1bee2p+121;
  xi[1] = -0x1.cb403ep-59;
  yr[1] = 0x1.480000p-144;
  yi[1] = -0x1.c66fc4p+5;
  zr[1] = -0x1.60b8cap-34;
  zi[1] = -0x1.e8b02ap+115;

  xr[2] = -0x1.3f6e00p-97;
  xi[2] = -0x1.c00000p-146;
  yr[2] = 0x1.000000p-148;
  yi[2] = -0x1.0c4e70p-91;
  zr[2] = 0x1.aa50d0p-55;
  zi[2] = -0x1.30c746p-6;

  xr[3] = 0x1.000000p-148;
  xi[3] = 0x1.f4bc04p-84;
  yr[3] = 0x1.00ad74p-20;
  yi[3] = 0x1.2ad02ep-85;
  zr[3] = 0x1.1102ccp-127;
  zi[3] = 0x1.f369a4p-64;

  for (i = 0; i < 4; i++)
    {
      if (debug)
	{
	  printf("ar  =%-16.6a ai   =%-16.6a\n",xr[i],xi[i]);
	  printf("br  =%-16.6a bi   =%-16.6a\n",yr[i],yi[i]);
	}
      __real__ a = xr[i];
      __imag__ a = xi[i];
      __real__ b = yr[i];
      __imag__ b = yi[i];
      __real__ z = zr[i];
      __imag__ z = zi[i];
      c = cdivf(a,b);
      cr = __real__ c;
      ci = __imag__ c;

      if (!match(c,z)){
	ok = 0;
      }
    }
  if (!ok)
    abort();
  exit(0);
}
