/*
  Build with gcc -lm -O2 chkd_cdiv.c cdivd_t?.c -o chkd_t?
  Program to test complex divide 
  for correct results on selected values.
  Checking known failure points.
*/

#include <math.h>
#include <complex.h>
#include <stdio.h>
extern void abort();
extern void exit();
extern long double complex cdivld(long double complex a,long double complex b);

#define SMALL (0x1.0p-16384L)
#define MAXBIT 63
#define ERRLIM 6
int debug = 0;

/*
  Compare c (computed value) with z (expected value)
  Return 0 if within allowed range
  Return 1 if not
*/
int match(complex c, complex z)
{
  long double rz, iz, rc, ic;
  long double rerr, ierr, rmax;
  long double biterr;
  rz = __real__ z;
  iz = __imag__ z;
  rc = __real__ c;
  ic = __imag__ c;
  if (isnan(rc) || isnan(ic)) return 1;

  if (fabs(rz) > SMALL)
    {
      rerr = fabs(rz-rc)/fabs(rz);
    }
  else if (fabs(rz) == 0.0)
    {
      rerr = fabs(rc);
    }
  else
    {
      rerr = fabs(rz-rc)/SMALL;
    }

  if (fabs(iz) > SMALL)
    {
      ierr = fabs(iz-ic)/fabs(iz);
    }
  else if (fabs(iz) == 0.0)
    {
      ierr = fabs(ic);
    }
  else
    {
      ierr = fabs(iz-ic)/SMALL;
    }
  rmax = fmax(rerr, ierr);
  biterr = 0;
  if ( rmax != 0.0)      
    {
      biterr = log2(rmax) + MAXBIT + 1;
    }
  if (debug)
      {
	/* debug */
	printf("biterr =%13.1f\n", biterr);
	printf("expect  zr %-24.13a zi   =%-24.13a\n",rz,iz);
	printf("observe cr %-24.13a ci   =%-24.13a\n\n",rc,ic);
      }

  if (biterr >= ERRLIM)
    return 0;
  else
    return 1;
}


int main(int argc, char** argv)
{
  long double complex a,b,c,z;
  long double xr[4], xi[4], yr[4], yi[4], zr[4], zi[4];
  long double cr, ci;
  int i;
  int ok = 1;
  xr[0] = -0x9.c793985b7d029d90p-8480L;
  xi[0] = 0x8.018745ffa61a8fe0p+16329L;
  yr[0] = -0xe.d5bee9c523a35ad0p-15599L;
  yi[0] = -0xa.8c93c5a4f94128f0p+869L;
  zr[0] = -0xc.248bc228e01adcb0p+15456L;
  zi[0] = -0x8.89baf6960fac5cf0p-1011L;

  xr[1] = 0xb.68e44bc6d0b91a30p+16026L;
  xi[1] = 0xb.ab10f5453e972f30p-14239L;
  yr[1] = 0x8.8cbd470705428ff0p-16350L;
  yi[1] = -0xa.0c1cbeae4e4b69f0p+347L;
  zr[1] = 0xf.76204243c2f28070p-1022L;
  zi[1] = 0x9.15b5abcc93d1f920p+15676L;

  xr[2] = -0x9.e8c093a43b546a90p+15983L;
  xi[2] = 0xc.95b18274208311e0p-2840L;
  yr[2] = -0x8.dedb729b5c1b2ec0p+8L;
  yi[2] = 0xa.a49fb81b24738370p-16385L;
  zr[2] = 0x8.efccf744dd88c7a0p+15972L;
  zi[2] = 0xa.b8f3f7c826b5b780p-421L;

  xr[3] = 0xc.4687f251c0f48bd0p-3940L;
  xi[3] = -0xe.a3f2138992d85fa0p+15598L;
  yr[3] = 0xe.4b0c25c3d5ebb830p-16344L;
  yi[3] = -0xa.6cbf1ba80f7b97a0p+78L;
  zr[3] = 0xb.3c2dd11dfdba2670p+15517L;
  zi[3] = -0xf.6773dc63dee9b110p-905L;


  for (i = 0; i < 4; i++)
    {
      if (debug)
	{
	  printf("ar  =%-24.13a ai   =%-24.13a\n",xr[i],xi[i]);
	  printf("br  =%-24.13a bi   =%-24.13a\n",yr[i],yi[i]);
	}
      __real__ a = xr[i];
      __imag__ a = xi[i];
      __real__ b = yr[i];
      __imag__ b = yi[i];
      __real__ z = zr[i];
      __imag__ z = zi[i];
      c = cdivld(a,b);
      cr = __real__ c;
      ci = __imag__ c;

      if (!match(c,z)){
	ok = 0;
      }
    }
  if (!ok)
    abort();
  exit(0);
}
