/* Test program for cdiv random number generation.
   Build with
   cc -lm rantestd.c mkrand.c cdivd_t?.c -O2 -o rantestd
   for full range of random numbers (exponent -1023 to + 1024)
   or
   cc -lm rantestd.c mkrand_small.c cdivd_t?.c -O2 -o rantestds
   for smaller range (exponent -512 to +512).
*/

#include <stdio.h>
#include <math.h>
#include <complex.h>

#define RUNLEN 10000000
#define MAXBIT 52
/* p-52 = 1 bit */
/* p-49 = 4 bit */
/* p-47 = 6 bit */
/* p-45 = 8 bit */
/* p-37 = 16 bit */
/* p-29 = 24 bit */
/* p-1  = 52 bit */
#define DIF 10
#define ERRLIMIT 50

#define ERRLIM (0x1.0p-1)
#define SMALL (0x1.0p-1024)


extern double mkrand(double);
double complex cdivd(double complex, double complex);
double infp, infm;

int debug = 0;


/* Return true if x/y can be computed in double precision.  */
int zzvalid(double complex a)
{
  if (debug){
    printf("valid = %-24.13a %-24.13ai\n",__real__ a,__imag__ a);
  }
  return ((__real__ a != infp) && (__real__ a != infm) &&
	  (__imag__ a != infp) && (__imag__ a != infm) );
}

/* Return number of bits x, y differ by.  */
double zzdiff(double complex a, double complex b)
{
  double ra,rb,rerr;
  double ia,ib,ierr;
  ra = __real__ a;
  ia = __imag__ a;
  rb = __real__ b;
  ib = __imag__ b;
  if (debug){
    printf("quad  = %-24.13a %-24.13ai\n",__real__ a,__imag__ a);
  }
  /* If new values include nan, then return error.  */
  if (isnan(rb) || isnan(ib)) return 999.0;
  if (debug){
    printf("test  = %-24.13a %-24.13ai\n",__real__ b,__imag__ b);
  }

  if (fabs(ra) > SMALL)
    {
      rerr = fabs(ra-rb)/fabs(ra);
    } 
  else if (fabs(ra) == 0.0)
    {
      rerr = fabs(rb);
    }
  else
    {
      rerr = fabs(ra-rb)/SMALL;
      if (debug) {
	printf("rr = %-24.13a %-24.13a %-24.13a\n",ra,rb,rerr);
      }
    }
  if (fabs(ia) > SMALL)
    {
      ierr = fabs(ia-ib)/fabs(ia);
    }
  else if (fabs(ia) == 0.0)
    {
      ierr = fabs(ib);
    }
  else
    {
      ierr = fabs(ia-ib)/SMALL;
      if (debug) {
	printf("ir = %-24.13a %-24.13a %-24.13a\n",ia,ib,ierr);
      }
    }

  if (debug){
    printf("err  = %-24.13a %-24.13ai\n",rerr,ierr);
    printf("rerr chk %d   ierr chk %d\n", (rerr > ERRLIM),(ierr > ERRLIM));
  }

  return fmax(rerr,ierr);
}


int main(int argc, char** argv)
{
  int bittbl[MAXBIT];
  int bitindx;
  long int i,cnt,dif;
  double biterr, zdif;
  
  /* Print the first errlimit values which exceed DIF.  */
  int errlimit = ERRLIMIT;
  double z;
  double complex x, y, ans, new;
  long double complex xx, yy, qq;
  infp = 1.0/0.0;
  infm = -1.0/0.0;
  z = 1.0;
  cnt = 0;
  dif = 0;
  for (i = 0; i < MAXBIT; i++) bittbl[i] = 0;

  while (cnt < RUNLEN)
    {
      __real__ x = mkrand(z);
      __imag__ x = mkrand(z);
      __real__ y = mkrand(z);
      __imag__ y = mkrand(z);
      xx = x;
      yy = y;
      qq = xx/yy;
      ans = qq;
      new = cdivd(x,y);
      if (zzvalid(ans))
	{
	  cnt++;
	  zdif = zzdiff(ans,new);
	  biterr = 0;
	  if (zdif != 0.0)
	    {
	      biterr = log2(zdif) + MAXBIT + 1;
	    }
	  bitindx = (int)biterr;
	  if (bitindx < 0) bitindx = 0;
	  if (bitindx > MAXBIT) bitindx = MAXBIT;
	  for ( i = 0; i < bitindx; i++ ) {
	    bittbl[i]++;
	  }
	  if (bitindx >= DIF)
	    {
	      if (debug) {
		double a,b,c,d,r,t;
		double ac,bd,bc,ad, acbd,bcad;
		a = __real__ x; b = __imag__ x; c = __real__ y; d = __imag__ y;
		if (fabs(c) > fabs(d)) {
		  r = d/c;
		  ac = a; bd = b*r; bc= b; ad=a*r;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c+r*d);
		} else {
		  r = c/d;
		  ac = a*r; bd = b; bc= b*r; ad=a;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c*r+d);
		}
		printf("ac   =%-24.13a bd   =%-24.13a\n",ac,bd);
		printf("bc   =%-24.13a ad   =%-24.13a\n",bc,ad);
		printf("ac+bd=%-24.13a bc-ad=%-24.13a\n",acbd,bcad);
		printf("r    =%-24.13a t    =%-24.13a\n",r,t);
	      }

	      dif++;
	      if (dif < errlimit) {
		printf("  x %-24.13a %-24.13a\n",__real__ x,__imag__ x);
		printf("  y %-24.13a %-24.13a\n",__real__ y,__imag__ y);
		printf("good %-24.13a %-24.13a\n",__real__ ans,__imag__ ans);
		printf("new  %-24.13a %-24.13a  biterr= %13.1f\n\n",
		       __real__ new,__imag__ new, biterr);
	      }
	    }
	}
    }
  printf("cnt= %9d  dif=%9d\n",cnt,dif);
  for (i = 0; i < MAXBIT; i++) 
    {
      printf("%2dbit= %7d\n",i+1,bittbl[i]);
    }
}
