/* Test program for cdivh random number generation.
   Build with
   cc -lm rantesth.c mkranh.c cdivh_t1.c -O2 -o rantest
   for full range of random numbers (exponent -126 to + 127)
   or
   cc -lm rantesth.c mkranh_small.c cdivh_t1.c -O2 -o rantesth
   for smaller range (exponent -63 to +64).
*/

#include <stdio.h>
#include <math.h>
#include <complex.h>

#define RUNLEN 1000000
#define MAXBIT 12
/* p-24 = 1 bit */  
/* p-23 = 2 bit */  
/* p-21 = 4 bit */  
/* p-17 = 8 bit */  
/* p-13 = 12 bit */
/* p-1 = 24 bit */  
#define DIF 5
#define ERRLIM (0x1.0p-1)
#define SMALL (0x1.0p-9)
#define ERRLIMIT 50

extern _Float16 mkranh(_Float16);
_Float16 complex cdivh(_Float16 complex, _Float16 complex);
_Float16 infp, infm;

int debug = 0;


/* Return true if x/y can be computed in _Float16 precision.  */
int zzvalid(_Float16 complex a)
{
  if (debug){
    printf("valid = %-13.3a %-13.3ai\n",__real__ a,__imag__ a);
  }
  return ((__real__ a != infp) && (__real__ a != infm) &&
	  (__imag__ a != infp) && (__imag__ a != infm) );
}

/* Return number of bits x, y differ by.  */
float zzdiff(_Float16 complex a, _Float16 complex b)
{
  float ra,rb,rerr;
  float ia,ib,ierr;
  float t1,t2;
  ra = __real__ a;
  ia = __imag__ a;
  rb = __real__ b;
  ib = __imag__ b;
  if (debug){
    printf("quad  = %-13.3a %-13.3ai\n",__real__ a,__imag__ a);
  }
  /* If new values include nan, then return error.  */
  if (isnan(rb) || isnan(ib)) return 99.0;
  if (debug){
    printf("test  = %-13.3a %-13.3ai\n",__real__ b,__imag__ b);
  }

  if (fabs(ra) > SMALL)
    {
      rerr = fabs(ra-rb)/fabs(ra);
    }
  else if (fabs(ra) == 0.0)
    {
      rerr = fabs(rb);
    }
  else
    {
      rerr = fabs(ra-rb)/SMALL;
      if (debug) {
	printf("rr = %-13.3a %-13.3a %-13.3a\n",ra,rb,rerr);
      }
    }
  if (fabs(ia) > SMALL)
    {
      ierr = fabs(ia-ib)/fabs(ia);
    }
  else if (fabs(ia) == 0.0)
    {
      ierr = fabs(ib);
    }
  else
    {
      ierr = fabs(ra-rb)/SMALL;
      if (debug) {
	printf("ir = %-13.3a %-13.3a %-13.3a\n",ia,ib,ierr);
      }
    }

  if (debug){
    printf("err  = %-13.3a %-13.3ai\n",rerr,ierr);
    printf("rerr chk %d   ierr chk %d\n", (rerr > ERRLIM),(ierr > ERRLIM));
  }

  return fmax(rerr,ierr);
}


int main(int argc, char** argv)
{
  int bittbl[MAXBIT];
  int bitindx;
  long int i,cnt,dif;
  float biterr;
  
  /* Print the first errlimit values which fail.  */
  int errlimit = ERRLIMIT;
  _Float16 z;
  _Float16 complex x, y, ans, new;
  long double complex xx, yy, qq;
  float xxr, xxi, yyr, yyi, ansr, ansi, newr, newi;
  infp = 1.0/0.0;
  infm = -1.0/0.0;
  z = 1.0;
  cnt = 0;
  dif = 0;
  for (i = 0; i < MAXBIT; i++) bittbl[i] = 0;

  while (cnt < RUNLEN)
    {
      __real__ x = mkranh(z);
      __imag__ x = mkranh(z);
      __real__ y = mkranh(z);
      __imag__ y = mkranh(z);
      xxr = __real__ x;
      xxi = __imag__ x;
      yyr = __real__ y;
      yyi = __imag__ y;
      xx = x;
      yy = y;
      qq = xx/yy;
      ans = qq;
      ansr = __real__ ans;
      ansi = __imag__ ans;
      new = cdivh(x,y);
      newr = __real__ new;
      newi = __imag__ new;
      if (zzvalid(ans))
	{
	  cnt++;
	  biterr = log2f(zzdiff(ans,new))+MAXBIT+1;
	  bitindx = (int)biterr;
	  if (bitindx < 0) bitindx = 0;
	  if (bitindx > MAXBIT) bitindx = MAXBIT;
	  for ( i = 0; i < bitindx; i++ ) {
	    bittbl[i]++;
	  }
	  if (biterr >= DIF)
	    {
	      if (debug) {
		_Float16 a,b,c,d,r,t;
		_Float16 ac,bd,bc,ad, acbd,bcad;
		a = __real__ x; b = __imag__ x; c = __real__ y; d = __imag__ y;
		if (fabs(c) > fabs(d)) {
		  r = d/c;
		  ac = a; bd = b*r; bc= b; ad=a*r;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c+r*d);
		} else {
		  r = c/d;
		  ac = a*r; bd = b; bc= b*r; ad=a;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c*r+d);
		}
		printf("ac   =%-13.3a bd   =%-13.3a\n",ac,bd);
		printf("bc   =%-13.3a ad   =%-13.3a\n",bc,ad);
		printf("ac+bd=%-13.3a bc-ad=%-13.3a\n",acbd,bcad);
		printf("r    =%-13.3a t    =%-13.3a\n",r,t);
	      }

	      dif++;
	      if (dif < errlimit) {
		printf("  x %-13.3a %-13.3a\n",xxr,xxi);
		printf("  y %-13.3a %-13.3a\n",yyr,yyi);
		printf("good %-13.3a %-13.3a\n",ansr,ansi);
		printf("new  %-13.3a %-13.3a  biterr= %13.1f\n\n",
		       newr,newi, biterr);
	      }
	    }
	}
    }
  printf("cnt= %9d  dif=%9d\n",cnt,dif);
  for (i = 0; i < MAXBIT; i++) 
    {
      printf("%2dbit= %7d\n",i+1,bittbl[i]);
    }
}
