/* test program for cdivf random number generation
   build with
   cc -lm rantestf.c mkranf.c cdivf_b0.c -O2 -o rantestf
   for full range of random numbers (exponent -128 to + 127)
   or
   cc -lm rantestf.c mkranf_small.c cdivf_b0.c -O2 -o rantestf
   for smaller range (exponent -63 to +64)
*/

#include <stdio.h>
#include <math.h>
#include <complex.h>

#define RUNLEN 200000000
#define MAXBIT 24
/* p-24 = 1 bit */  
/* p-23 = 2 bit */  
/* p-21 = 4 bit */  
/* p-17 = 8 bit */  
/* p-13 = 12 bit */
/* p-1 = 24 bit */  
#define DIF 10
#define ERRLIMIT 50

#define ERRLIM (0x1.0p-1)
#define SMALL (0x1.0p-128)



extern float mkranf(float);
float complex cdivf(float complex, float complex);
float infp, infm;

int debug = 0;


/* Return true if x/y can be computed in float precision.  */
int zzvalid(float complex a)
{
  if (debug){
    printf("valid = %-16.6a %-16.6ai\n",__real__ a,__imag__ a);
  }
  return ((__real__ a != infp) && (__real__ a != infm) &&
	  (__imag__ a != infp) && (__imag__ a != infm) );
}

/* Return number of bits x, y differ by.  */
float zzdiff(float complex a, float complex b)
{
  float ra,rb,rerr;
  float ia,ib,ierr;
  float t1,t2;
  ra = __real__ a;
  ia = __imag__ a;
  rb = __real__ b;
  ib = __imag__ b;
  if (debug){
    printf("quad  = %-16.6a %-16.6ai\n",__real__ a,__imag__ a);
  }
  /* If new values include nan, then return error.  */
  if (isnan(rb) || isnan(ib)) return 999.0;
  if (debug){
    printf("test  = %-16.6a %-16.6ai\n",__real__ b,__imag__ b);
  }

  if (fabsf(ra) > SMALL)
    {
      rerr = fabsf(ra-rb)/fabsf(ra);
    } 
  else if (fabsf(ra) == 0.0)
    {
      rerr = fabsf(rb);
    }
  else
    {
      rerr = fabs(ra-rb)/SMALL;
      if (debug) {
	printf("rr = %-16.6a %-16.6a %-16.6a\n",ra,rb,rerr);
      }
    }
  if (fabsf(ia) > SMALL)
    {
      ierr = fabsf(ia-ib)/fabsf(ia);
    }
  else if (fabsf(ia) == 0.0)
    {
      ierr = fabsf(ib);
    }
  else
    {
      ierr = fabsf(ia-ib)/SMALL;
      if (debug) {
	printf("ir = %-16.6a %-16.6a %-16.6a\n",ia,ib,ierr);
      }
    }

  if (debug){
    printf("err  = %-16.6a %-16.6ai\n",rerr,ierr);
    printf("rerr chk %d   ierr chk %d\n", (rerr > ERRLIM),(ierr > ERRLIM));
  }

  return fmaxf(rerr,ierr);
}


int main(int argc, char** argv)
{
  int bittbl[MAXBIT];
  int bitindx;
  long int i,cnt,dif;
  float biterr, zdif;
  
  /* Print the first errlimit values which exceed DIF.  */
  int errlimit = ERRLIMIT;
  float z;
  float complex x, y, ans, new;
  long double complex xx, yy, qq;
  infp = 1.0/0.0;
  infm = -1.0/0.0;
  z = 1.0;
  cnt = 0;
  dif = 0;
  for (i = 0; i < MAXBIT; i++) bittbl[i] = 0;

  while (cnt < RUNLEN)
    {
      __real__ x = mkranf(z);
      __imag__ x = mkranf(z);
      __real__ y = mkranf(z);
      __imag__ y = mkranf(z);
      xx = x;
      yy = y;
      qq = xx/yy;
      ans = qq;
      new = cdivf(x,y);
      if (zzvalid(ans))
	{
	  cnt++;
	  zdif = zzdiff(ans,new);
	  biterr = 0.0;
	  if (zdif != 0)
	    {
	      biterr = log2f(zdif) + MAXBIT + 1;
	    }
	  bitindx = (int)biterr;
	  if (bitindx < 0) bitindx = 0;
	  if (bitindx > MAXBIT) bitindx = MAXBIT;
	  for ( i = 0; i < bitindx; i++ ) {
	    bittbl[i]++;
	  }
	  if (bitindx >= DIF)
	    {
	      if (debug) {
		float a,b,c,d,r,t;
		float ac,bd,bc,ad, acbd,bcad;
		a = __real__ x; b = __imag__ x; c = __real__ y; d = __imag__ y;
		if (fabsf(c) > fabsf(d)) {
		  r = d/c;
		  ac = a; bd = b*r; bc= b; ad=a*r;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c+r*d);
		} else {
		  r = c/d;
		  ac = a*r; bd = b; bc= b*r; ad=a;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c*r+d);
		}
		printf("ac   =%-16.6a bd   =%-16.6a\n",ac,bd);
		printf("bc   =%-16.6a ad   =%-16.6a\n",bc,ad);
		printf("ac+bd=%-16.6a bc-ad=%-16.6a\n",acbd,bcad);
		printf("r    =%-16.6a t    =%-16.6a\n",r,t);
	      }

	      dif++;
	      if (dif < errlimit) {
		printf("  x %-16.6a %-16.6a\n",__real__ x,__imag__ x);
		printf("  y %-16.6a %-16.6a\n",__real__ y,__imag__ y);
		printf("good %-16.6a %-16.6a\n",__real__ ans,__imag__ ans);
		printf("new  %-16.6a %-16.6a  biterr= %13.1f\n\n",
		       __real__ new,__imag__ new, biterr);
	      }
	    }
	}
    }
  printf("cnt= %9d  dif=%9d\n",cnt,dif);
  for (i = 0; i < MAXBIT; i++) 
    {
      printf("%2dbit= %7d\n",i+1,bittbl[i]);
    }
}
