/* Test program for cdivx random number generation.
   Build with
   cc -lm rantestx.c mkranx.c cdivx_t?.c -O2 -o rantestx
   for full range of random numbers (exponent -16384 to + 16383)
   or
   cc -lm rantestx.c mkranx_small.c cdivx_t?.c -O2 -o rantestxs
   for smaller range (exponent -8192 to +8191).
*/

#include <stdio.h>
#include <math.h>
#include <complex.h>

#define RUNLEN 10000000
#define MAXBIT 63
/* p-63 = 1 bit */  
/* p-62 = 2 bit */  
/* p-60 = 4 bit */  
/* p-56 = 8 bit */  
/* p-52 = 12 bit */
/* p-1  = 63 bit */  
#define DIF 10
#define ERRLIMIT 50

#define ERRLIM (0x1.0p-1)
#define SMALL1 (0x1.0p-1024)
#define SMALL2 (0x1.0p-16)


extern long double mkranx(long double);
long double complex cdivx(long double complex, long double complex);
long double infp, infm;

int debug = 0;


/* Return true if x/y can be computed in long double precision.  */
int zzvalid(long double complex a)
{
  if (debug){
    printf("valid = %-27.16La %-27.16Lai\n",__real__ a,__imag__ a);
  }
  return ((__real__ a != infp) && (__real__ a != infm) &&
	  (__imag__ a != infp) && (__imag__ a != infm) );
}

/* Return number of bits x, y differ by.  */
long double zzdiff(long double complex a, long double complex b)
{
  long double ra,rb,rerr;
  long double ia,ib,ierr;
  long double t1,t2;
  long double small = (long double)SMALL1 * (long double)SMALL2;
  ra = __real__ a;
  ia = __imag__ a;
  rb = __real__ b;
  ib = __imag__ b;
  if (debug){
    printf("quad  = %-27.16La %-27.16Lai\n",ra,ia);
  }
  /* If new values include nan, then return error.  */
  if (isnanl(rb) || isnanl(ib)) return (long double)999.0;
  if (debug){
    printf("test  = %-27.16La %-27.16Lai\n",rb,ib);
  }

  if (fabsl(ra) > small)
    {
      rerr = fabsl(ra-rb)/fabsl(ra);
    } 
  else if (fabsl(ra) == 0.0)
    {
      rerr = fabsl(rb);
    }
  else
    {
      rerr = fabsl(ra-rb)/small;
      if (debug) {
	printf("rr = %-27.16La %-27.16La %-27.16La\n",ra,rb,rerr);
      }
    }
  if (fabsl(ia) > small)
    {
      ierr = fabsl(ia-ib)/fabsl(ia);
    }
  else if (fabsl(ia) == 0.0)
    {
      ierr = fabsl(ib);
    }
  else
    {
      ierr = fabsl(ia-ib)/small;
      if (debug) {
	printf("ir = %-27.16La %-27.16La %-27.16La\n",ia,ib,ierr);
      }
    }

  if (debug){
    printf("err  = %-27.16La %-27.16Lai\n",rerr,ierr);
    printf("rerr chk %d   ierr chk %d\n", (rerr > ERRLIM),(ierr > ERRLIM));
  }

  return fmaxl(rerr,ierr);
}


int main(int argc, char** argv)
{
  int bittbl[MAXBIT];
  int bitindx;
  long int i,cnt,dif;
  long double biterr, zdif;
  
  /* Print the first errlimit values which exceed DIF.  */
  int errlimit = ERRLIMIT;
  long double z;
  long double complex x, y, ans, new;
  long double complex xx, yy, qq;
  infp = 1.0/0.0;
  infm = -1.0/0.0;
  z = 1.0;
  cnt = 0;
  dif = 0;
  for (i = 0; i < MAXBIT; i++) bittbl[i] = 0;

  while (cnt < RUNLEN)
    {
      __real__ x = mkranx(z);
      __imag__ x = mkranx(z);
      __real__ y = mkranx(z);
      __imag__ y = mkranx(z);
      xx = x;
      yy = y;
      qq = xx/yy;
      ans = qq;
      new = cdivx(x,y);
      if (zzvalid(ans))
	{
	  cnt++;
	  zdif = zzdiff(ans, new);
	  biterr = 0;
	  if (zdif != 0.0)
	    {
	      biterr = log2l(zdif) + MAXBIT + 1;
	    }
	  bitindx = (int)biterr;
	  if (bitindx < 0) bitindx = 0;
	  if (bitindx > MAXBIT) bitindx = MAXBIT;
	  for ( i = 0; i < bitindx; i++ ) {
	    bittbl[i]++;
	  }

	  if (bitindx >= DIF)
	    {
	      if (debug) {
		long double a,b,c,d,r,t;
		long double ac,bd,bc,ad, acbd,bcad;
		a = __real__ x; b = __imag__ x; c = __real__ y; d = __imag__ y;
		if (fabsl(c) > fabsl(d)) {
		  r = d/c;
		  ac = a; bd = b*r; bc= b; ad=a*r;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c+r*d);
		} else {
		  r = c/d;
		  ac = a*r; bd = b; bc= b*r; ad=a;
		  acbd = ac+bd;
		  bcad = bc-ad;
		  t = 1.0/(c*r+d);
		}
		printf("ac   =%-27.16La bd   =%-27.16La\n",ac,bd);
		printf("bc   =%-27.16La ad   =%-27.16La\n",bc,ad);
		printf("ac+bd=%-27.16La bc-ad=%-27.16La\n",acbd,bcad);
		printf("r    =%-27.16La t    =%-27.16La\n",r,t);
	      }

	      dif++;
	      if (dif < errlimit) {
		printf("  x %-27.16La %-27.16La\n",__real__ x,__imag__ x);
		printf("  y %-27.16La %-27.16La\n",__real__ y,__imag__ y);
		printf("good %-27.16La %-27.16La\n",__real__ ans,__imag__ ans);
		printf("new  %-27.16La %-27.16La biterr= %11.1Lf\n\n",
		       __real__ new,__imag__ new, biterr);
	      }
	    }
	}
    }
  printf("cnt= %9d  dif=%9d\n",cnt,dif);
  for (i = 0; i < DIF; i ++)
    {
      printf("%2dbit= %7d\n",i+1,bittbl[i]);
    }
  for (i = DIF; i < MAXBIT; i++) 
    {
      if (bittbl[i] != 0) printf("%2dbit= %7d\n",i+1,bittbl[i]);
    }
}
